
(define gnc:version "2.6.19")

;; Automatically generated defaults (don't use these directly --
;; they're used during actual initialization elsewhere)
(define gnc:_install-doc-path_ '("/usr/share"))
